// To compile on Linux:
// cc -Wall MandelbrotOGL.cpp -o MandelbrotOGL -lGL -lGLU -lglut -lGLEW -lm -lX11 -L /usr/X11R6/lib
// To compile on OSX:
// g++ -framework GLUT -framework OpenGL MandelbrotOGL.cpp -o MandelbrotOGL
// Mandelbrot Client
// Original CUDA program submitted by Mark Granger, NewTek
// Client was written by Abdulmajed Dakkak

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef __APPLE__
    #include <GLUT/glut.h>
#else
    #include <GL/glut.h>
#endif
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#define IMAGE_W		800
#define IMAGE_H		600
#define MAX_PASS    4

#define INVALIDATE_STEP 3

#define GETNEWFRAME 0x0
#define ZOOMIN      0x2
#define ZOOMOUT     0x3
#define MOVERIGHT   0x4
#define MOVEUP      0x5
#define MOVELEFT    0x6
#define MOVEDOWN    0x7
#define TERMINATE   0x8
#define MOUSE       0x9
#define CLICK       0x1
#define MOTION      0x2
#define LEFTCLICK   0x0
#define RIGHTCLICK  0x1

//Original image width and height
int imageW, imageH, max_pass;

// Starting position and scale
double xOff = -0.5;
double yOff = 0.0;
double scale = 3.2;

// Starting stationary position and scale motion
double xdOff = 0.0;
double ydOff = 0.0;

// SOCKET ID
int sockfd;

// Request packet
char request[16];

// D_DST Def
char d_dst[IMAGE_H * IMAGE_W * 4];

short invalid_frame = true;

/**********Network Functions************/
void connectToServer(int port, char* IP) {
   fprintf(stderr,"Attempting to connect to server...\n");
   struct sockaddr_in remote_addr;

   sockfd=socket(PF_INET,SOCK_STREAM,0);
   
   remote_addr.sin_family=AF_INET;
   remote_addr.sin_port=htons(port);
   remote_addr.sin_addr.s_addr=inet_addr(IP);
   memset(remote_addr.sin_zero, '\0',sizeof remote_addr.sin_zero);

   if(connect(sockfd,(struct sockaddr*)&remote_addr,sizeof remote_addr)==-1){
      perror("connect");
      exit(1);
   }
   
   fprintf(stderr,"Connected to server.\n");
}

void sendEnvironmentVariables( ) {
    int env[3] = {imageW, imageH, MAX_PASS};
    send(sockfd, env, 3 * sizeof(int), 0);
}

void getNewFrame( ) {
//   fprintf(stderr,"Requesting new frame...\n");

   request[0]= GETNEWFRAME;
   send(sockfd, request, 16 * sizeof(char), 0);

   int bytesRecieved=recv(sockfd, d_dst, 
                          imageW * imageH * 4 * sizeof(char), MSG_WAITALL);
   if(bytesRecieved==-1) { perror("recv()"); exit(1); }

//   fprintf(stderr,"Received frame. (%d bytes).\n",bytesRecieved);
}

void sendRequestToZoomIn( ) {
    request[0] = ZOOMIN;
    send(sockfd, request, 16 * sizeof(char), 0);
}

void sendRequestToZoomOut( ) {
    request[0] = ZOOMOUT;
    send(sockfd, request, 16 * sizeof(char), 0);
}

void sendRequestToMoveLeft( ) {
    request[0] = MOVELEFT;
    send(sockfd, request, 16 * sizeof(char), 0);
}

void sendRequestToMoveUp( ) {
    request[0] = MOVEUP;
    send(sockfd, request, 16 * sizeof(char), 0);
}

void sendRequestToMoveRight( ) {
    request[0] = MOVERIGHT;
    send(sockfd, request, 16 * sizeof(char), 0);
}

void sendRequestToMoveDown( ) {
    request[0] = MOVEDOWN;
    send(sockfd, request, 16 * sizeof(char), 0);
}

void sendRequestToTerminate( ) {
    request[0] = TERMINATE;
    send(sockfd, request, 16 * sizeof(char), 0);
}

void sendClickEvent(int button_clicked, int x, int y) {
    short button;
    if(button_clicked == 0) {
        button = LEFTCLICK;
    }
    else {
        button = RIGHTCLICK;
    }

    request[0] = MOUSE;
    request[1] = CLICK;
    request[2] = button;
    int x_y[2] = {x, y};
    memcpy(&request[3], x_y, 2*sizeof(int));
//    printf("Sent Click event %d, %d, %d\n", button, x_y[0], x_y[1]);
    send(sockfd, (void*)request, 16 * sizeof(char), 0);
}

void sendMotionEvent(int x, int y) {
    request[0] = MOUSE;
    request[1] = MOTION;

    int x_y[2] = {x, y};
    memcpy(&request[2], x_y, 2*sizeof(int));

    send(sockfd, request, 16 * sizeof(char), 0);
}

/*****END**Network Functions**END*******/

// OpenGL display function
void displayFunc(void)
{
    int i, j;
    char r,g,b,a;
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glColor3f(0.0, 0.0, 0.0);
    glBegin(GL_POINTS);
    for(j = 0; j < imageH; j++) {
        for(i = 0; i < imageW; i++) {
            r = d_dst[j*4*imageW + i * 4];
            g = d_dst[j*4*imageW + i * 4 + 1];
            b = d_dst[j*4*imageW + i * 4 + 2];
            a = d_dst[j*4*imageW + i * 4 + 3];
            glColor4b(r, g, b, a);
            glVertex2f(i, j);
        }
    }
    glEnd( );
	glutSwapBuffers();
} // displayFunc

// OpenGL keyboard function
void keyboardFunc(unsigned char k, int, int)
{
    invalid_frame = true;
    switch (k){
        case '\033':
        case 'q':
        case 'Q':
            printf("Sending server terminate request...\n");
            sendRequestToTerminate( );
            printf("Shutdown done.\n");
            exit(0);
            break;

        case '?':
            printf("xOff = %5.8f\n", xOff);
            printf("yOff = %5.8f\n", yOff);
            printf("scale = %5.8f\n", scale);
            printf("\n");
            break;
        
        case '4':	// Left arrow key
			xOff -= 0.05f * scale;
            sendRequestToMoveLeft( );
		    break;
         
        case '8':	// Up arrow key
			yOff += 0.05f * scale;
            sendRequestToMoveUp( );
		    break;
        
        case '6':	// Right arrow key
			xOff += 0.05f * scale;
            sendRequestToMoveRight( );
		    break;
         
        case '2':	// Down arrow key
			yOff -= 0.05f * scale;
            sendRequestToMoveDown( );
		    break;
		
        case '+':
			scale /= 1.1f;
            sendRequestToZoomIn( );
		    break;
         
        case '-':
			scale *= 1.1f;
            sendRequestToZoomOut( );
		    break;
		
		default:
		    break;
   }
} // keyboardFunc

// OpenGL mouse click function
void clickFunc(int button, int, int x, int y)
{
    sendClickEvent(button, x, y);
    invalid_frame = true;
} // clickFunc

// OpenGL mouse motion function
void motionFunc(int x, int y)
{
    sendMotionEvent(x, y);
    invalid_frame = true;
} // motionFunc

void idleFunc()
{
    if(invalid_frame) {
        getNewFrame( );
        glutPostRedisplay();
        invalid_frame++;
        if(invalid_frame == INVALIDATE_STEP) {
            invalid_frame = 0;
        }
    }
}


void init_gl( )
{
    glClearColor(0.0, 0.0, 0.0, 0.0);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0.0, imageW, 0.0, imageH);

    glMatrixMode(GL_MODELVIEW);
    glPointSize(2);
}

////////////////////////////////////////////////////////////////////////////////
// Main program
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
    if(argc!=3) {
        fprintf(stderr,"Usage: %s <IP address> <port>\n",argv[0]);
        exit(1);
    }


	imageW = IMAGE_W;
	imageH = IMAGE_H;
    max_pass = MAX_PASS;

    connectToServer(atoi(argv[2]),argv[1]);
    sendEnvironmentVariables( );

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
    glutInitWindowSize(imageW, imageH);
    glutCreateWindow(argv[0]);

    init_gl( );
    glutDisplayFunc(displayFunc);
    glutIdleFunc(idleFunc);
    glutKeyboardFunc(keyboardFunc);
    glutMouseFunc(clickFunc);
    glutMotionFunc(motionFunc);
    glutMainLoop();

} // main
